/*
  xphd - X Pointer Hiding Daemon

  xphd.c: main routine

  Copyright (C) 2003  Yusuke Shinyama <yusuke at cs . nyu . edu>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  http://www.gnu.org/licenses/gpl.html
*/

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <unistd.h> 
#include <stdio.h> 
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>

#define FIXED_FONT "fixed"
#define EMPTY_CHAR ' '
#define IGNORE_IF_KEYSYM_LOWERBOUND 0xFFE1	/* ignore modifier keys */

#define MOTION_THRESHOLD 2
#define REDISPLAY_INTERVAL 60

Display* disp;
Window root;
Window current_focus;
Cursor empty_cursor;
int x_root0, y_root0;


void Bind1(Window w)
{
    XSelectInput(disp, w, KeyPressMask | PointerMotionMask | SubstructureNotifyMask);
}

void Unbind1(Window w)
{
    XSelectInput(disp, w, 0);
}

void BindAll(Window w)
{
    Window root_return, parent_return, *children_return;
    unsigned int nchildren_return, i;
    Bind1(w);
    if (XQueryTree(disp, w, &root_return, &parent_return, &children_return, &nchildren_return)) {
	for(i = 0; i < nchildren_return; i++) {
	    BindAll(children_return[i]);
	}
	XFree((void*) children_return);
    }
}

void InitCursor(void)
{
    Font font_fixed;
    XColor color_black = { 0, 0, 0, 0, 0, 0 };

    font_fixed = XLoadFont(disp, FIXED_FONT);
    empty_cursor = XCreateGlyphCursor(disp, font_fixed, font_fixed, 
				      EMPTY_CHAR, EMPTY_CHAR, 
				      &color_black, &color_black);
}

/* something very very dirty... */
static int err_flag;
typedef int (*Handler)(Display*, XErrorEvent*);
static Handler error_handler0;	/* original handler */
static int ignore_badwindow(Display* d, XErrorEvent* ev)
{
    if (ev->error_code != BadWindow) {
	/* process the error with the original handler */
	return error_handler0(d, ev);
    }
    err_flag = 1;
    return 0;
}

/* find the lowest window */
Window FindWindow(void)
{
    Window w, root_return, child_return;
    int root_x_return, root_y_return;
    int win_x_return, win_y_return;
    unsigned int mask_return;
    w = root;
    err_flag = 0;
    while(!err_flag) {
	(void)XQueryPointer(disp, w,
			    &root_return, &child_return,
			    &root_x_return, &root_y_return,
			    &win_x_return, &win_y_return,
			    &mask_return);
	if (child_return == None || err_flag) break;
	w = child_return;
    }
    return w;
}

void HideCursor(void)
{
    if (current_focus == None) {
	current_focus = FindWindow();
#ifdef DEBUG
	fprintf(stderr, "hide: 0x%08lx\n", (unsigned long)current_focus);
#endif
	XDefineCursor(disp, current_focus, empty_cursor);
    }
}

void ShowCursor(void)
{
    if (current_focus != None) {
#ifdef DEBUG
	fprintf(stderr, "show: 0x%08lx\n", (unsigned long)current_focus);
#endif
	XUndefineCursor(disp, current_focus);
	current_focus = None;
    }
}

/* signal handlers */
void Cleanup(int signum)
{
    ShowCursor();
    XCloseDisplay(disp);
    exit(0);
}

void ShowAgain(int signum)
{
    ShowCursor();
    XFlush(disp);
}

/* main */
int main(int argc, char* argv[])
{
    disp = XOpenDisplay(NULL);
    root = DefaultRootWindow(disp);
    current_focus = None;

    /* ignoring BadWindow error, this is necessary because 
     */
    error_handler0 = XSetErrorHandler(ignore_badwindow);

    InitCursor();
    BindAll(root);

    signal(SIGINT, Cleanup);
    signal(SIGHUP, Cleanup);
    signal(SIGQUIT, Cleanup);
    signal(SIGTERM, Cleanup);

    signal(SIGALRM, ShowAgain);

    /* event loop */
    while(1) {
	XEvent event_return;
	XNextEvent(disp, &event_return);
	switch(event_return.type) {
	case KeyPress:
	    /* ignore modifier keys themselves */
	    if (IGNORE_IF_KEYSYM_LOWERBOUND <= XLookupKeysym(&event_return.xkey, 0))
		break;
	    HideCursor();
	    x_root0 = event_return.xkey.x_root;
	    y_root0 = event_return.xkey.y_root;
	    alarm(REDISPLAY_INTERVAL);
	    break;
	case MotionNotify:
	    if (MOTION_THRESHOLD <= abs(x_root0 - event_return.xmotion.x_root) ||
		MOTION_THRESHOLD <= abs(y_root0 - event_return.xmotion.y_root)){ 
		ShowCursor();
	    }
	    break;
	case MappingNotify:	/* mandatory for all X11 apps. */
	    XRefreshKeyboardMapping(&event_return.xmapping);
	    break;
	case MapNotify:
#ifdef DEBUG
	    fprintf(stderr, "mapped: 0x%08lx\n", (unsigned long)event_return.xmap.window);
#endif
	    if (current_focus != None && current_focus != FindWindow()) {
		ShowCursor();
	    }
	    BindAll(root);
	    break;
	case UnmapNotify:
#ifdef DEBUG
	    fprintf(stderr, "unmapped: 0x%08lx\n", (unsigned long)event_return.xmap.window);
#endif
	    if (current_focus != None && current_focus != FindWindow()) {
#ifdef DEBUG
		fprintf(stderr, "cleared: 0x%08lx\n", (unsigned long)current_focus);
#endif
		current_focus = None;
	    }
	    break;
	default:
	    break;
	}
    }
    /*NOTREACHED*/
    return 0;
}
